require "operations/railtie"
require "operations/utils"
require "operations/core_ext"
require "operations/engine"
require "operations/config"
require "operations/enforcer"
require "operations/act_as_operationable"
require "operations/errors"
require "operations/operation"
require "operations/version"

module Operations
  class << self
    def operations_list
      Operations::Config.get_operations_list
    end

    def user_roles
      Operations::Config.user_roles
    end

    def named_user_roles(all: false)
      roles = Operations::Config.user_roles
      roles = roles.reject{|role| role[:db_value].nil?} unless all
      roles.map{|role| role[:name]}
    end

    def operations_list_by_uuid
      operations_list.map{|op| {operation: op, uuid: op.uuid}}
    end

    # Returns the operation associated with the passed in UUID
    def from_uuid(uuid)
      operations_list.select{|operation| operation.uuid == uuid}[0]
    end

    # Returns the operation associated with the passed in name
    def from_string(name)
      operations_list.select{|operation| operation.name.to_s == name.to_s}[0]
    end

    # Returns the operations that have scope scope
    def allowing(scope)
      return nil if scope.blank?
      operations_list.select{|operation| operation > scope}
    end

    def allowed_named_roles_for(scope)
      current_value = user_role_value_from(scope)
      user_roles
          .select{|role| role[:name] == :all ||
              !role[:value].nil? && role[:value] <= current_value}
          .map{|role| role[:name]}
    end

    def from_user_type(scope)
      value = allowed_named_roles_for(scope.to_sym)
      scopes = value.reject{|role| role.nil?}
    end

    def allowed_int_roles_for(scope)
      allowed_named_roles_for(scope)
          .map{|named_scope| user_role_int_from(named_scope)}
    end

    def allowed_value_roles_for(scope)
      allowed_named_roles_for(scope)
          .map{|named_scope| user_role_value_from(named_scope)}
    end

    # Determines if the user is allowed to execute the operation name
    def allows?(user, name)
      operation = from_string(name)
      return false if operation.nil?
      operation.users.map{|u| u.id}.include?(user.id)
    end

    # Returns a list of user roles names straight from the config file
    def user_roles_names
      user_roles.map{|ur| ur[:name]}
    end

    # Returns a list of users roles that qualify as role_name
    def user_roles_from(role_name, select_only: false)
      value = user_role_value_from(role_name)
      return [] if value.nil?
      compare =-> (role, value) do
        select_only ? role[:value] == value : role[:value] <= value
      end
      user_roles
          .select{|role| role[:value] && compare.call(role, value)}
          .map{|role| role[:db_value]}
          .reject{|res| res.nil?}
    end

    # Returns a list of users roles that do not qualify as role_name
    def user_roles_until(role_name)
      value = user_role_value_from(role_name)
      return [] if value.nil?
      user_roles
          .select{|role| role[:value] && role[:value] > value}
          .map{|role| role[:db_value]}
    end

    def scopes_as(scope)
      value = user_role_value_from(scope)
      return [] if value.nil?
      return [scope] if %w{all nobody}.include?(scope.to_s)
      user_roles
          .select{|role| role[:value] && role[:value] >= value}
          .map{|role| role[:name]}
    end

    def operations_for(scope)
      scopes = scopes_as(scope)
      list = operations_list
      if %w{all nobody}.include?(scope.to_s)
        list.select{|op| op == scope}.uniq
      else
        list.select{|op| op >= scope}.uniq
      end
    end

    # Returns a list of users that have roles user_roles_from
    def users_acting_as(role_name, select_only: false)
      User.where(role: user_roles_from(role_name, select_only: select_only))
    end

    # Returns a list of users that have roles user_roles_until
    def users_denied_from(role_name)
      User.where(role: user_roles_until(role_name))
    end

    def user_role_int_from(role_name)
      result = user_roles.select{|role| role[:name] == role_name}[0]
      return nil if result.nil?
      result[:db_value]
    end

    def user_role_value_from(role_name)
      result = user_roles.select{|role| role[:name] == role_name}[0]
      return nil if result.nil?
      result[:value]
    end

    def user_role_name_from(role_int)
      result = user_roles.select{|role| role[:db_value] == role_int}[0]
      return nil if result.nil?
      result[:name]
    end
  end
end
