module Operations
  class Operation
    OperationValue = Struct.new :name, :scope, :description

    def initialize(**options, &block)
      if block_given?
        yield values
      else
        options.each do |key, value|
          if values.respond_to?(key)
            values.send(key.to_s + '=', value)
          end
        end
      end
      ensure_operation_is_valid!
      self
    end

    def values
      @values ||= OperationValue.new(nil, :nobody)
    end

    def name
      has_valid_name? ? values.name : 'N/A (Invalid operation name)'
    end

    def description
      values.description
    end

    def users
      return User.where(id: []) if scope == :nobody
      return User.all if scope == :all
      @users ||= Operations.users_acting_as(scope)
    end

    # def user
    #   return nil if values.user_id.nil?
    #   @user ||= User.find_by(id: values.user_id)
    # end

    # def save
    #   return false if user.nil?
    #   ensure_operation_is_valid!("Cannot save user #{user}")
    #   result = user.update_attribute :operations, self.to_json
    # end

    def is_valid?
      has_valid_name? && has_valid_scope? # && !user.nil?
    end

    def has_valid_name?
      Operations::Config.operation_name_regex === values.name
    end

    def has_valid_scope?
      Operations::Config.operation_scope_regex === scope
    end

    def invalid_scope
      return scope unless has_valid_scope?
    end

    def is_scope?(scope_name)
      scope.to_s == scope_name.to_s
    end

    def accepts_scope?(scope_name)
      scope_name = scope_name.to_sym
      return true if scope == :all
      return false if scope == :nobody
      return true if is_scope?(scope_name)
      Operations.allowed_named_roles_for(scope).include?(scope_name)
    end

    def allowed_roles
      Operations.user_roles
    end

    def denied_roles
    end

    def as_json(options={})
      additional_hash = {uuid: uuid, users_count: users.count}
      if (methods = options[:methods])
        if methods.class != Array
          methods = [methods]
        end
        methods.each do |method|
          begin
            additional_hash[method] = send(method)
          rescue NoMethodError
            # ignore
            next
          end
        end
      end
      values
          .as_json(options)
          .merge(additional_hash)
          .as_json
    end

    def uuid
      alg = Operations::Config.operation_uuid_algorithm
      return Operations::Errors::BaseException("The UUID Algorithm has be a class") unless alg.class == Class
      alg.new(name.to_s).to_s
    end

    def to_json
      as_json.to_json
    end

    def inspect
      "\#<#{self.class.name} `#{self.name}' allowed for `#{scope}'>"
    end

    def ==(sc_op)
      if sc_op.class == self.class
        sc_op.name == name && sc_op.uuid == uuid
      else
        is_scope?(sc_op)
      end
    end

    def >(sc_op)
      if sc_op.class == self.class
        return false if self == sc_op
        return !sc_op.accepts_scope?(scope)
      elsif sc_op.class == Symbol
        accepts_scope?(sc_op)
      else
        false
      end
    end

    def <(sc_op)
      if sc_op.class == self.class
        return false if self == sc_op
        return sc_op.accepts_scope?(scope)
      elsif sc_op.class == Symbol
        !accepts_scope?(sc_op)
      else
        false
      end
    end

    def >=(sc_op)
      self == sc_op || self > sc_op
    end

    def <=(sc_op)
      self == sc_op || self < sc_op
    end

    private

    def ensure_operation_is_valid!(msg=nil)
      raise Operations::Errors::InvalidOperationError.new(self, nil) unless is_valid?
    end

    # Hide the scope for security reasons
    def scope
      self.values.scope || :nobody
    end
  end
end
